package org.exoplatform.services.cms.metadata;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.Session;
import javax.jcr.Value;
import javax.jcr.nodetype.NodeType;
import javax.jcr.nodetype.NodeTypeIterator;
import javax.jcr.nodetype.PropertyDefinition;

import org.exoplatform.container.ExoContainer;
import org.exoplatform.container.ExoContainerContext;
import org.exoplatform.container.component.ComponentPlugin;
import org.exoplatform.container.configuration.ConfigurationManager;
import org.exoplatform.container.xml.PortalContainerInfo;
import org.exoplatform.services.cache.CacheService;
import org.exoplatform.services.cache.ExoCache;
import org.exoplatform.services.cms.BasePath;
import org.exoplatform.services.cms.CmsConfigurationService;
import org.exoplatform.services.cms.impl.Utils;
import org.exoplatform.services.cms.templates.impl.TemplateConfig;
import org.exoplatform.services.cms.templates.impl.TemplatePlugin;
import org.exoplatform.services.jcr.RepositoryService;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeType;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager;
import org.exoplatform.services.templates.velocity.impl.JCRResourceLoaderImpl;
import org.picocontainer.Startable;

/**
 * @author Hung Nguyen Quang
 * @mail   nguyenkequanghung@yahoo.com
 */

public class MetadataServiceImpl implements MetadataService, Startable{

  static final public String METADATA_TYPE = "exo:metadata".intern() ;
  static final public String NT_UNSTRUCTURED = "nt:unstructured".intern() ;
  final static public String INTERNAL_USE = "exo:internalUse".intern() ;
  static final public String EXO_TEMPLATE = "exo:template".intern() ;
  static final public String EXO_ROLES_PROP = "exo:roles".intern() ;
  static final public String EXO_TEMPLATE_FILE_PROP = "exo:templateFile".intern() ;
  static final public String DIALOGS = "dialogs".intern() ;
  static final public String VIEWS = "views".intern() ;
  static final public String DIALOG1 = "dialog1".intern() ;
  static final public String VIEW1 = "view1".intern() ;

  private RepositoryService repositoryService_;
  private CmsConfigurationService cmsConfigService_ ;
  private String metadataTemplatePath_ ;
  private List<TemplatePlugin> plugins_ = new ArrayList<TemplatePlugin>();

  public MetadataServiceImpl(CmsConfigurationService cmsConfigService, RepositoryService repositoryService) throws Exception{
    cmsConfigService_ = cmsConfigService ;
    repositoryService_ = repositoryService ;    
    metadataTemplatePath_ = cmsConfigService_.getJcrPath(BasePath.METADATA_PATH) ;
  }

  public void addPlugins(ComponentPlugin plugin) {
    if (plugin instanceof TemplatePlugin)
      plugins_.add((TemplatePlugin) plugin);    
  }

  public void addMetadataTemplate(String nodetype, boolean isDialog, String[] role, String content, boolean isAddNew) throws Exception {        
    Session session = repositoryService_.getRepository().getSystemSession(cmsConfigService_.getWorkspace()) ;    
    Node metadataHome = (Node)session.getItem(metadataTemplatePath_) ;    
    if(!isAddNew) {
      if(isDialog){
        Node dialog1 = metadataHome.getNode(nodetype).getNode(DIALOGS).getNode(DIALOG1) ;
        dialog1.setProperty(EXO_ROLES_PROP, role);
        dialog1.setProperty(EXO_TEMPLATE_FILE_PROP, content);
        dialog1.save() ;        
        removeFromCache(dialog1.getPath());
      }else {
        Node view1 = metadataHome.getNode(nodetype).getNode(VIEWS).getNode(VIEW1) ;
        view1.setProperty(EXO_ROLES_PROP, role);
        view1.setProperty(EXO_TEMPLATE_FILE_PROP, content);
        view1.save() ;
        removeFromCache(view1.getPath()) ;
      }      
    }else {
      Node metadata = null ;
      if(metadataHome.hasNode(nodetype)) {
        metadata = metadataHome.getNode(nodetype) ;
      }else {
        metadata = metadataHome.addNode(nodetype, NT_UNSTRUCTURED) ;
      }
      addTemplate(metadata, role, content, isDialog) ;
      metadataHome.save() ;
    }           
    session.save() ;    
    session.logout();
  }

  public void removeMetadataTemplate(String metadataType) throws Exception {
    String dialogPath = getMetadataTemplatePath(metadataType,true) ;
    String viewPath = getMetadataTemplatePath(metadataType,false) ;
    Session session = repositoryService_.getRepository().getSystemSession(cmsConfigService_.getWorkspace()) ;    
    Node metadataHome = (Node)session.getItem(metadataTemplatePath_) ;
    Node metadata = metadataHome.getNode(metadataType) ; 
    metadata.remove() ;
    metadataHome.save() ;
    session.save() ;
    session.logout() ;
    removeFromCache(dialogPath) ;
    removeFromCache(viewPath) ;
  } 

  public String getMetadataTemplate(String name, boolean isDialog) throws Exception {    
    Node template = getMetadataTemplateNode(name,isDialog) ;
    if(template == null) return null ;
    return template.getProperty(EXO_TEMPLATE_FILE_PROP).getString();
  }

  public String getMetadataTemplatePath(String name, boolean isDialog) throws Exception {    
    Node template = getMetadataTemplateNode(name,isDialog) ;
    if(template == null) return null ;
    return template.getPath();
  }

  public String[] getMetadataTemplateRoles(String name, boolean isDialog) throws Exception {
    List<String> roleList = new ArrayList<String>() ;
    Node template = getMetadataTemplateNode(name,isDialog) ;
    if(template == null) return null ;
    for(Value role: template.getProperty(EXO_ROLES_PROP).getValues()) {
      roleList.add(role.getString()) ;
    }
    return roleList.toArray(new String[roleList.size()]) ;
  }  

  public List<String> getAllMetadataTypes() throws Exception {
    List<String> metadataTypes = new ArrayList<String>() ;
    for(NodeType metadata:getAllMetadataNodeType()) 
      metadataTypes.add(metadata.getName()) ;
    return metadataTypes ;
  }

  public NodeType getMetadataNodeType(String metadataTypeName) throws Exception {       
    NodeType metadataType = repositoryService_.getRepository().getNodeTypeManager().getNodeType(metadataTypeName) ;
    if(metadataType.isNodeType(METADATA_TYPE)) 
      return metadataType ;
    return null ;    
  }

  public List<String> getExtenalMetadataType() throws Exception {
    List<String> extenalMetaTypes = new ArrayList<String>() ;
    for(NodeType metadata: getAllMetadataNodeType()) {      
      ExtendedNodeType extNT = (ExtendedNodeType)metadata ;
      PropertyDefinition internalUseDef = extNT.getPropertyDefinitions(INTERNAL_USE).getAnyDefinition() ;
      if(!internalUseDef.getDefaultValues()[0].getBoolean()) 
        extenalMetaTypes.add(metadata.getName()) ;
    }
    return extenalMetaTypes ;
  }

  private List<NodeType> getAllMetadataNodeType() throws Exception {
    List<NodeType> metadataTypes = new ArrayList<NodeType>() ;    
    ExtendedNodeTypeManager ntManager = repositoryService_.getRepository().getNodeTypeManager();     
    for(NodeTypeIterator ntIter = ntManager.getMixinNodeTypes();ntIter.hasNext();) {
      NodeType nt = ntIter.nextNodeType() ;
      if(nt.isNodeType(METADATA_TYPE)) {
        metadataTypes.add(nt) ;
      }
    }
    return metadataTypes ;
  }
  private Node createTemplateHome(Node nodetype, boolean isDialog) throws Exception{
    if(isDialog) {
      Node dialogs = null ;
      if(nodetype.hasNode(DIALOGS)) {
        dialogs = nodetype.getNode(DIALOGS) ;      
      }else {
        dialogs = nodetype.addNode(DIALOGS, NT_UNSTRUCTURED) ;
      }    
      return dialogs ;
    }
    Node views = null ;
    if(nodetype.hasNode(VIEWS)) {
      views = nodetype.getNode(VIEWS) ;      
    }else {
      views = nodetype.addNode(VIEWS, NT_UNSTRUCTURED) ;
    }    
    return views ;    
  }

  private void addTemplate(Node nodetype, String[] role, String content, boolean isDialog) throws Exception {
    Node templateHome = createTemplateHome(nodetype, isDialog) ;
    Node template = null ;
    if(isDialog) {
      if(templateHome.hasNode(DIALOG1)) {
        template = templateHome.getNode(DIALOG1) ;        
      }else {
        template = templateHome.addNode(DIALOG1, EXO_TEMPLATE) ;
      }
    }else {
      if(templateHome.hasNode(VIEW1)) {
        template = templateHome.getNode(VIEW1) ;        
      }else {
        template = templateHome.addNode(VIEW1, EXO_TEMPLATE) ;
      }
    }    
    template.setProperty(EXO_ROLES_PROP, role) ;
    template.setProperty(EXO_TEMPLATE_FILE_PROP, content) ;    
  }

  private Node getMetadataTemplateNode(String name,boolean isDialog) throws Exception{
    Session session = repositoryService_.getRepository().getSystemSession(cmsConfigService_.getWorkspace()) ;    
    Node metadataHome = (Node)session.getItem(metadataTemplatePath_) ;
    String defaulDialog = new StringBuffer()
    .append(name).append("/").append(DIALOGS)
    .append("/").append(DIALOG1).toString();

    String defaulView = new StringBuffer()
    .append(name).append("/").append(VIEWS)
    .append("/").append(VIEW1).toString() ;
    session.logout();
    if(isDialog) {
      if(metadataHome.hasNode(defaulDialog))
        return metadataHome.getNode(defaulDialog) ;
      return null ;
    }         
    if(metadataHome.hasNode(defaulView))
      return metadataHome.getNode(defaulView) ;
    return null ;
  }

  public void start() {
    Session session = null ;
    try{      
      Node metadataHome = null ;
      List nodetypes = null;
      session = repositoryService_.getRepository().getSystemSession(cmsConfigService_.getWorkspace()) ;
      for (int j = 0; j < plugins_.size(); j++) {
        nodetypes = plugins_.get(j).getNodeTypes();
        if (nodetypes.isEmpty()){
          session.logout() ;
          return;
        }      
        TemplateConfig.NodeType nodeType = (TemplateConfig.NodeType) nodetypes.iterator().next();
        String nodeTypeName = nodeType.getNodetypeName();
        try {        
          metadataHome = (Node)session.getItem(metadataTemplatePath_);
        } catch (PathNotFoundException e) { }

        String sourcePath = cmsConfigService_.getContentLocation() + "/system" 
        + metadataTemplatePath_.substring(metadataTemplatePath_.lastIndexOf("/")) ;

        for (Iterator iter = nodetypes.iterator(); iter.hasNext();) {
          nodeType = (TemplateConfig.NodeType) iter.next();
          nodeTypeName = nodeType.getNodetypeName();
          Node nodeTypeHome = null;
          if (!metadataHome.hasNode(nodeTypeName)){
            nodeTypeHome = Utils.makePath(metadataHome, nodeTypeName, NT_UNSTRUCTURED);          
          } else{
            nodeTypeHome = metadataHome.getNode(nodeTypeName);
          }
          List dialogs = nodeType.getReferencedDialog();
          if(dialogs.size() > 0) {
            Node dialogsHome = createTemplateHome(nodeTypeHome, true);
            addNode(sourcePath, dialogsHome, dialogs);
          }
          List views = nodeType.getReferencedView();
          if(views.size() > 0) {
            Node viewsHome = createTemplateHome(nodeTypeHome, false);
            addNode(sourcePath, viewsHome, views);
          }
        }
      }    
      session.save();
      session.logout() ;
    }catch (Exception e) {
      if(session !=null) session.logout();
      e.printStackTrace() ;
    }    
  }

  public void stop() { }

  private void init() throws Exception{               
  }

  protected void removeFromCache(String templatePath) {
    ExoContainer container = ExoContainerContext.getCurrentContainer() ;
    CacheService cacheService = (CacheService)container.getComponentInstanceOfType(CacheService.class) ;
    PortalContainerInfo containerInfo = (PortalContainerInfo)container.getComponentInstanceOfType(PortalContainerInfo.class) ;
    try{
      ExoCache jcrcache_ = cacheService.getCacheInstance(JCRResourceLoaderImpl.class.getName()) ;
      String portalName = containerInfo.getContainerName() ;
      String key = portalName + "jcr:" +templatePath ; 
      Object cachedobject = jcrcache_.get(key);            
      if (cachedobject != null) {        
        jcrcache_.remove(key);      
      }
    }catch (Exception e) {    
      //e.printStackTrace();
    }       
  }  

  private void addNode(String metadataPath, Node templateHome, List templates)  throws Exception {    
    for (Iterator iterator = templates.iterator(); iterator.hasNext();) {
      TemplateConfig.Template template = (TemplateConfig.Template) iterator.next();
      String templateFileName = template.getTemplateFile();
      String path = metadataPath + templateFileName;
      ExoContainer container = ExoContainerContext.getCurrentContainer() ;
      ConfigurationManager configurationManager = 
        (ConfigurationManager)container.getComponentInstanceOfType(ConfigurationManager.class) ;
      InputStream in = configurationManager.getInputStream(path);
      String nodeName = 
        templateFileName.substring(templateFileName.lastIndexOf("/") + 1, templateFileName.indexOf("."));
      Node contentNode = templateHome.addNode(nodeName, EXO_TEMPLATE);
      contentNode.setProperty(EXO_ROLES_PROP, template.getParsedRoles());
      contentNode.setProperty(EXO_TEMPLATE_FILE_PROP, in);
    }
  }

}
