/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.InvalidSerializedDataException;
import javax.jcr.ItemExistsException;
import javax.jcr.ItemNotFoundException;
import javax.jcr.NamespaceRegistry;
import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.UnsupportedRepositoryOperationException;
import javax.jcr.Workspace;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.NodeTypeManager;
import javax.jcr.observation.ObservationManager;
import javax.jcr.query.QueryManager;
import javax.jcr.version.Version;
import javax.jcr.version.VersionException;
import javax.jcr.version.VersionHistory;

import org.apache.commons.logging.Log;
import org.exoplatform.container.ExoContainer;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.itemvisitors.NodeCloneVisitor;
import org.exoplatform.services.jcr.impl.core.itemvisitors.NodeCopyVisitor;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.core.query.QueryManagerFactory;
import org.exoplatform.services.jcr.impl.core.version.VersionImpl;
import org.exoplatform.services.jcr.impl.dataflow.session.WorkspaceStorageDataManagerProxy;
import org.exoplatform.services.jcr.impl.xml.NodeImporter;
import org.exoplatform.services.log.ExoLogger;
import org.xml.sax.ContentHandler;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: WorkspaceImpl.java 7541 2006-08-03 09:48:48Z peterit $
 */

public class WorkspaceImpl implements Workspace {

  protected static Log log = ExoLogger.getLogger("jcr.WorkspaceImpl");

  private SessionImpl session;

  private ExoContainer container;

  private WorkspaceStorageDataManagerProxy dataManager;

  private ObservationManager observationManager;

  private String name;

  public WorkspaceImpl(String name, ExoContainer container, SessionImpl session,
      ObservationManager observationManager) throws RepositoryException {

    this.session = session;
    this.container = container;
    this.dataManager = session.getTransientNodesManager().getWorkspaceDataManager();
//    this.dataManager = (BatcheableWorkspaceDataManager) container
//        .getComponentInstanceOfType(BatcheableWorkspaceDataManager.class);
    this.name = name;
    this.observationManager = observationManager;
  }

  /**
   * @see javax.jcr.Workspace#getName
   */
  public String getName() {
    return name;
  }

  /**
   * @see javax.jcr.Workspace#importXML TODO (the same as Session.importXML - ?)
   */
  public ContentHandler getImportContentHandler(String parentAbsPath, int uuidBehavior)
      throws PathNotFoundException, ConstraintViolationException, VersionException,
      RepositoryException {    
    NodeImporter contentHandler = (NodeImporter) session.getImportContentHandler(parentAbsPath, uuidBehavior);
    // Need to save all changes direct on SAX events
    contentHandler.setSaveOnEnd(true); 
    return contentHandler;
  }

  /**
   * @see javax.jcr.Workspace#importXML TODO (the same as Session.importXML - ?)
   */
  public void importXML(String parentAbsPath, InputStream in, int uuidBehavior) throws IOException,
      PathNotFoundException, ItemExistsException, ConstraintViolationException,
      InvalidSerializedDataException, RepositoryException {

    session.importXML(parentAbsPath, in, uuidBehavior);
    session.save();
  }

  /**
   * @see javax.jcr.Workspace#getSession
   */
  public Session getSession() {
    return session;
  }

  /**
   * @see javax.jcr.Workspace#getAccessibleWorkspaceNames
   */
  public String[] getAccessibleWorkspaceNames() throws RepositoryException {
    RepositoryImpl rep = (RepositoryImpl) session.getRepository();
    return rep.getWorkspaceNames();
  }

  /**
   * @see javax.jcr.Workspace#copy
   */
  public void copy(String srcWorkspace, String srcPath, String destPath)
      throws NoSuchWorkspaceException, ConstraintViolationException, VersionException,
      AccessDeniedException, PathNotFoundException, ItemExistsException, RepositoryException {

    copyNode(srcWorkspace, srcPath, destPath, false);
  }

  /**
   * @see javax.jcr.Workspace#copy
   */
  public void copy(String srcPath, String destPath) throws ItemExistsException, VersionException,
      PathNotFoundException, ItemExistsException, ConstraintViolationException, RepositoryException {

    copyNode(this.name, srcPath, destPath, false);
  }

  /**
   * @see javax.jcr.Workspace#move
   */
  public void move(String srcPath, String destPath) throws ConstraintViolationException,
      VersionException, AccessDeniedException, PathNotFoundException, ItemExistsException,
      RepositoryException {

    NodeImpl srcNode = (NodeImpl) getSession().getItem(srcPath);
    
    InternalQPath srcNodeQPath = srcNode.getLocation().getInternalPath();
    JCRPath destNodePath = session.getLocationFactory().parseAbsPath(destPath);
    InternalQPath destNodeQPath = session.getLocationFactory().parseAbsPath(destPath).getInternalPath();
    
    NodeImpl destNode = null;
    
    //----------- BEGIN SESSION BATCH ----------
    dataManager.beginSessionBatch(this.session);
    if (srcNode.isNodeType("mix:referenceable")) {
      destNode = moveNode(srcNode, destNodePath, true);
      session.getTransientNodesManager().addMovedPaths(srcNodeQPath, destNodeQPath);
    } else {
      // copyNode(this.name, srcPath, destPath, false);
      destNode = moveNode(srcNode, destNodePath, false);
      session.getTransientNodesManager().addMovedPaths(srcNodeQPath, destNodeQPath);
    }
    dataManager.endSessionBatch(this.session);
    //----------- END SESSION BATCH ----------
    
    // [PN] 06.04.06
    session.getTransientNodesManager().resetItemState(destNode);
    
    // clean session cache
    session.getTransientNodesManager().removeFromCache(srcNode);
    
    // clean moved storage
    session.getTransientNodesManager().removeMovedPathDescendants(srcNodeQPath);       
    session.getTransientNodesManager().removeMovedPathDescendants(destNodeQPath);
  }

  private NodeImpl moveNode(NodeImpl srcNode, JCRPath destNodePath, boolean keepUUIDs)
      throws NoSuchWorkspaceException, ConstraintViolationException, VersionException,
      AccessDeniedException, PathNotFoundException, ItemExistsException, RepositoryException {

    //  make dest node
    if (destNodePath.isIndexSetExplicitly())
      throw new RepositoryException("DestPath should not contain an index " + destNodePath.getAsString(true));

    String destParentPath = destNodePath.makeParentPath().getAsString(false);
    NodeImpl destParentNode = (NodeImpl) getSession().getItem(destParentPath);
    
    // Check if versionable ancestor is not checked-in
    if (!destParentNode.isCheckedOut())
      throw new VersionException("Destination parent node " + destParentNode.getPath()
          + " or its nearest ancestor is checked-in");
    if (!srcNode.isCheckedOut())
      throw new VersionException("Source parent node " + srcNode.getPath()
          + " or its nearest ancestor is checked-in");    
    
    // Check locking
    if (!destParentNode.checkLocking())
      throw new LockException("Destination parent node " + destParentNode.getPath() + " is locked ");
    if (!srcNode.checkLocking())
      throw new LockException("Source parent node " + srcNode.getPath() + " is locked ");
        
    //NodeImpl destNode = destParentNode.createChildNode(destNodePath.getName().getAsString(), srcNode
    //    .getPrimaryNodeType().getName(), srcNode.getMixinTypeNames(), true);
    
    NodeImpl destNode = destParentNode.copyChildNodeInmemory(
        destNodePath.getName().getInternalName(), 
        (NodeData)srcNode.getActualItemData(), keepUUIDs, false, false, false, false);

    //destNode.loadData(srcNode.getActualItemData());
    
    NodeCopyVisitor initializer = new NodeCopyVisitor(destNode, keepUUIDs, false);
    srcNode.accept(initializer);

    //dataManager.validate(destNode);
    
    // [PN] It's a magic:) To fix as fast as possible. To use DELETED status and saveItem
    // [PN] 09.04.06
    //dataManager.delete(srcNode);
    srcNode.removeFromParent();
    session.getTransientNodesManager().getTransactManager().saveItem(srcNode);
    
    //dataManager.add(destNode);
    session.getTransientNodesManager().getTransactManager().saveItem(destNode);
    return destNode; 
  }
  
  private void copyNode(String srcWorkspace, String srcPath, String destPath, boolean keepUUIDs)
      throws NoSuchWorkspaceException, ConstraintViolationException, VersionException,
      AccessDeniedException, PathNotFoundException, ItemExistsException, RepositoryException {

    SessionImpl srcSession = (SessionImpl) ((RepositoryImpl) session.getRepository()).login(session
        .getCredentials(), srcWorkspace);

    NodeImpl srcNode = (NodeImpl) srcSession.getItem(srcPath);
    
    //  make dest node
    JCRPath destNodePath = session.getLocationFactory().parseAbsPath(destPath);
    if (destNodePath.isIndexSetExplicitly())
      throw new RepositoryException("DestPath should not contain an index " + destPath);

    String destParentPath = destNodePath.makeParentPath().getAsString(false);
    NodeImpl destParentNode = (NodeImpl) getSession().getItem(destParentPath);
    
    // Check if versionable ancestor is not checked-in
    if (!destParentNode.isCheckedOut())
      throw new VersionException("Destination parent node " + destParentNode.getPath()
          + " or its nearest ancestor is checked-in");
    if (!srcNode.isCheckedOut())
      throw new VersionException("Source parent node " + srcNode.getPath()
          + " or its nearest ancestor is checked-in");    
    
    // Check locking
    if (!destParentNode.checkLocking())
      throw new LockException("Destination parent node " + destParentNode.getPath() + " is locked ");
    if (!srcNode.checkLocking())
      throw new LockException("Source parent node " + srcNode.getPath() + " is locked ");
        
    //----------- BEGIN SESSION BATCH ----------    
    dataManager.beginSessionBatch(this.session);
    
    NodeImpl destNode = destParentNode.createChildNode(destNodePath.getName().getAsString(), srcNode
        .getPrimaryNodeType().getName(), srcNode.getMixinTypeNames(), true);    
    NodeCopyVisitor initializer = new NodeCopyVisitor(destNode, keepUUIDs, false);
    srcNode.accept(initializer);
    //dataManager.validate(destNode);
    
    // [PN] 06.04.06
    //dataManager.add(destNode);
    session.getTransientNodesManager().getTransactManager().saveItem(destNode);
    dataManager.endSessionBatch(this.session);
    //----------- END SESSION BATCH ----------
    
    // [PN] 06.04.05
    session.getTransientNodesManager().resetItemState(destNode); 
  }

  /**
   * @see javax.jcr.Workspace#clone
   */
  public void clone(String srcWorkspace, String srcPath, String destPath, boolean removeExisting)
      throws NoSuchWorkspaceException, ConstraintViolationException, VersionException,
      AccessDeniedException, PathNotFoundException, ItemExistsException, RepositoryException {

    if (srcWorkspace.equals(this.name))
      throw new RepositoryException("Source and destination workspace are equals " + name);

    // make dest node path
    JCRPath destNodePath = session.getLocationFactory().parseAbsPath(destPath);
    if (destNodePath.isIndexSetExplicitly())
      throw new RepositoryException("DestPath should not contain an index " + destPath);

    // find src node
    SessionImpl srcSession = (SessionImpl) ((RepositoryImpl) session.getRepository()).login(session
        .getCredentials(), srcWorkspace);

    NodeImpl srcNode = (NodeImpl) srcSession.getItem(srcPath);

    // get parent node
    String parentPath = destNodePath.makeParentPath().getAsString(false);
    // check if there is a parent in persistence storage
//    if (dataManager.getItemData(destNodePath.makeParentPath()) == null)
    if (session.getTransientNodesManager().getTransactManager().getItemData(destNodePath.makeParentPath().getInternalPath()) == null)
      throw new PathNotFoundException("Destination parent node " + parentPath
          + " exists but is not stored in " + name);
    NodeImpl parentNode = (NodeImpl) getSession().getItem(parentPath);

    // Check if versionable ancestor is not checked-in
    if (!parentNode.isCheckedOut())
      throw new VersionException("Node " + parentNode.getPath()
          + " or its nearest ancestor is checked-in");

    // Check locking
    if (!parentNode.checkLocking())
      throw new LockException("Node " + parentNode.getPath() + " is locked ");

    // check for existed same nodes and remove or throw depending on
    // removeExisting flag
    String uuid = null;
    List nodesToRemove = new ArrayList();
    if (srcNode.isNodeType("mix:referenceable")) {
      uuid = srcNode.getUUID();
      // check root by uuid
      try {
        NodeImpl relNode = (NodeImpl) session.getNodeByUUID(uuid);
        if (!removeExisting) {
          throw new ItemExistsException("Item exists uuid = " + uuid + " at "
              + session.getWorkspace().getName() + " path: " + relNode.getPath());
        } else {
          nodesToRemove.add(relNode);
        }
      } catch (ItemNotFoundException e) { // save UUID (i.e. clone)
        log.debug("Destination referenceable node with uuid '" + uuid + "' not found" + e);
      }
    }

    //----------- BEGIN SESSION BATCH ----------
    dataManager.beginSessionBatch(session);
    removeNodesList(nodesToRemove);
    NodeImpl destNode = parentNode.copyChildNodeInmemory(destNodePath.getInternalPath().getName(),
        (NodeData) srcNode.getActualItemData(), srcNode.isNodeType("mix:referenceable"), false,
        false, false, false);
    // initialize dest node's children
    NodeCloneVisitor initializer = new NodeCloneVisitor(destNode, removeExisting, nodesToRemove);
    srcNode.accept(initializer);
    // remove existed same nodes (if removeExisting = true)
    removeNodesList(nodesToRemove);
    // add cloned node
    //dataManager.validate(destNode);
    
    // [PN]
    //dataManager.add(destNode);
    session.getTransientNodesManager().getTransactManager().saveItem(destNode);
    dataManager.endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    
    // [PN] 06.04.05
    session.getTransientNodesManager().resetItemState(destNode); 
    
    // [PN] Here we will mark cloned node parent to reload data from workspace, 
    // and do same for all destination node descendants in session cache
    parentNode.setRefreshToPersistedVersion(true);
    session.getTransientNodesManager().refreshCache(parentNode.getLocation());
    //session.getTransientNodesManager().removeFromCache(parentNode);
  }

  private int removeNodesList(List nodesToRemove) throws ConstraintViolationException,
      RepositoryException {
    // remove existed same nodes (if removeExisting = true)
    int removed = nodesToRemove.size();
    for (Iterator it = nodesToRemove.iterator(); it.hasNext();) {
      Object o = it.next();
      if (o instanceof NodeImpl) {
        NodeImpl node = (NodeImpl) o;
        // TODO it is possible the node is already removed along with removed
        // parent
        try {
          // dataManager.delete(node);
          node.remove();          
          node.save();
          //System.err.println("====== node removed: " + node.getPath() + " uuid: " + node.getInternalUUID());
        } catch (ItemExistsException e) {
          // item allready removed
          //System.err.println("====== item allready removed: " + e);
        }
      }
    }
    nodesToRemove.clear();
    return removed;
  }

  /**
   * @see javax.jcr.Workspace#getQueryManager
   */
  public QueryManager getQueryManager() throws RepositoryException {
    QueryManagerFactory qf = (QueryManagerFactory) container
        .getComponentInstanceOfType(QueryManagerFactory.class);

    if (qf == null)
      throw new RepositoryException("Query Manager Factory not found. Check configuration.");
    else
      return qf.getQueryManager(session);
  }

  /**
   * @see javax.jcr.Workspace#getNodeTypeManager
   */
  public NodeTypeManager getNodeTypeManager() throws RepositoryException {
    //System.out.println("SESSSSSSSSSSSSS "+session);
    return ((NodeTypeManagerImpl) container.getComponentInstanceOfType(NodeTypeManager.class))
    .createWorkspaceNTManager(session);
  }

  /**
   * @see javax.jcr.Workspace#getNamespaceRegistry
   */
  public NamespaceRegistry getNamespaceRegistry() throws RepositoryException {
    return (NamespaceRegistry) container.getComponentInstanceOfType(NamespaceRegistry.class);
  }

  /**
   * @see javax.jcr.Workspace#getObservationManager
   */
  public ObservationManager getObservationManager() throws UnsupportedRepositoryOperationException,
      RepositoryException {
    return observationManager;
  }

  private class MultiVersionRestoreComparator implements Comparator<Version> {
    public int compare(Version v1, Version v2) {      
      return 0;
    }
  }  
    
  /**
   * @see javax.jcr.Workspace#restore
   */
  public void restore(Version[] versions, boolean removeExisting)
      throws UnsupportedRepositoryOperationException, VersionException, RepositoryException,
      InvalidItemStateException {

    if (session.hasPendingChanges())
      throw new InvalidItemStateException("Session has pending changes ");

    // [PN] 20.02.06 We don't use these vars, as we will throw exception in for-cycle
    // check if there are existed node, for verification
    // List<JCRPath> notExisted = new ArrayList<JCRPath>();
    // List<JCRPath> existed = new ArrayList<JCRPath>();          
    
    Exception versionException = null;
    InternalQPath versionStorage = InternalQPath.parse(Constants.JCR_VERSION_STORAGE_URI);
    
    // for restore operation
    List<String> existedUuids = new ArrayList<String>(); //InWorkspace
    List<Version> notExistedVersions = new ArrayList<Version>();    
    LinkedHashMap<Version, NodeImpl> existedVersions = new LinkedHashMap<Version, NodeImpl>();

    for (Version v: versions) {          
      VersionHistory vh = v.getContainingHistory();
      String versionableUuid = vh.getVersionableUUID();
      NodeImpl node = null;
      try {
        node = (NodeImpl) session.getNodeByUUID(versionableUuid);
        if(node.getActualItemData().getQPath().isDescendantOf(versionStorage, false)) {
          continue; // skip version storage nodes (TODO:Remove in future)
        }
        // log.info("=== restore(): version: " + v.getPath() + " " + node.getPath() 
        //   + " " + node.getProperty("jcr:uuid").getString() + " " + versionableUuid);
        existedVersions.put(v, node);
        existedUuids.add(versionableUuid);
      } catch (ItemNotFoundException e) {        
        // not found, looking for parent
        // SPEC: For every version V in S that corresponds to a missing node in the workspace, there must
        // also be a parent of V in S
        // =========================================
        // [PN], 14.02.06, So, we will play for TCK and will look at spec.
        // Trying search for corresponding node, we hope that it's in VersionImpl object
        // If we have a corr node and her parent in the existed list - all ok, otherwise exception will throws.
        NodeImpl corrNode = null;
        String versionableParentUuid = null;  
        if (!v.getSession().getWorkspace().getName().equals(session.getWorkspace().getName())) {
          try {
            corrNode = (NodeImpl) v.getSession().getNodeByUUID(versionableUuid);
            if(corrNode.getActualItemData().getQPath().isDescendantOf(versionStorage, false)) {
              continue; // skip version storage nodes (TODO:Remove in future)
            }  
            versionableParentUuid = corrNode.getParent().getProperty("jcr:uuid").getString();
            // log.info("=== restore(): has versionable corr node: " + v.getSession().getWorkspace().getName() + ":" 
            //   + v.getPath() + ")  corr node:" + corrNode.getPath());
          } catch (ItemNotFoundException e1) {
            // not found in version session
            log.warn("Workspace.restore(): corr node not found: " + versionableUuid, e);
            versionException = e;
          } catch (RepositoryException e1) {
            log.warn("Workspace.restore(): Repository error, for corr node: " + versionableUuid, e);
            versionException = e;
          }
        }        
        if (versionableParentUuid != null && existedUuids.contains(versionableParentUuid)) {
          notExistedVersions.add(v);
          continue;
        } else {
          throw new VersionException("No such node (for version, from the array of versions, "
            + "that corresponds to a missing node in the workspace, " 
            + "there must also be a parent in the array). UUID: " + versionableUuid, versionException);
        }
      }
      // [PN] 20.02.06
//      if (node.isNew()) {
//        notExisted.add(node.getLocation());
//      } else {
//        existed.add(node.getLocation());
//      }
    }
//    recognizeExistense(notExisted, existed);
//    if (notExisted.size() > 0) {
//      throw new VersionException("There are not existed nodes with not existed ancestors in set "
//          + notExisted.get(0));
//    }

    for (Version v: existedVersions.keySet()) {
      String versionableUuid = v.getContainingHistory().getVersionableUUID();      
      // [PN], 14.02.06, 20.02.06      
      try {
        //NodeImpl node = (NodeImpl) session.getNodeByUUID(versionableUuid);
        NodeImpl node = existedVersions.get(v);
        //log.warn("Workspace.restore(): " + node.getPath() + "  " + node.getUUID() + "  " + removeExisting);
        ((VersionImpl) v).restore(node, removeExisting);          
      } catch (ItemNotFoundException e) {
        throw new VersionException("No such node. UUID: " + versionableUuid, e);
      } catch (ItemExistsException e) {
        log.warn("Workspace.restore(): " + e.getMessage());
        throw e;
      } catch (RepositoryException e) {
        log.warn("Workspace.restore(): Repository error: " + versionableUuid, e);
        throw e;
      }
    }
    for (Version v : notExistedVersions) {
      String versionableUuid = v.getContainingHistory().getVersionableUUID();
      NodeImpl node = null;
      try {
        node = (NodeImpl) session.getNodeByUUID(versionableUuid);     
        ((VersionImpl) v).restore(node, removeExisting);
      } catch (ItemNotFoundException e) {
        throw new VersionException("No such node restored before (for version, from the array of versions, "
            + "that corresponds to a missing node in the workspace, "
            + "there must also be a parent in the array). UUID: " + versionableUuid);
      } catch (ItemExistsException e) {
        log.warn("Workspace.restore() for item not existed before: " + e.getMessage());
        throw e;
      } catch (RepositoryException e) {
        log.warn("Workspace.restore() for item not existed before: Repository error: " + versionableUuid, e);
        throw e;
      }
    } 
  }
}