/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core;

import java.io.IOException;
import java.io.InputStream;
import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.InvalidSerializedDataException;
import javax.jcr.ItemExistsException;
import javax.jcr.NamespaceRegistry;
import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.UnsupportedRepositoryOperationException;
import javax.jcr.Workspace;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.NodeTypeManager;
import javax.jcr.observation.ObservationManager;
import javax.jcr.query.QueryManager;
import javax.jcr.version.Version;
import javax.jcr.version.VersionException;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.exoplatform.container.ExoContainer;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeType;
import org.exoplatform.services.jcr.dataflow.DataManager;
import org.exoplatform.services.jcr.dataflow.ItemDataChangesLog;
import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.core.query.QueryManagerFactory;
import org.exoplatform.services.jcr.impl.core.version.VersionImpl;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataCloneVisitor;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataCopyVisitor;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataMoveVisitor;
import org.exoplatform.services.jcr.impl.dataflow.MutableItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.session.SessionChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.version.VersionHistoryDataHelper;
import org.exoplatform.services.jcr.impl.xml.NodeImporter;
import org.exoplatform.services.log.ExoLogger;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: WorkspaceImpl.java 11887 2007-01-15 15:16:12Z peterit $
 */

public class WorkspaceImpl implements Workspace {

  protected static Log       log = ExoLogger.getLogger("jcr.WorkspaceImpl");

  private SessionImpl        session;

  private ExoContainer       container;

  // private WorkspaceStorageDataManagerProxy dataManager;

  private ObservationManager observationManager;

  private String             name;

  public WorkspaceImpl(String name, ExoContainer container, SessionImpl session,
      ObservationManager observationManager) throws RepositoryException {

    this.session = session;
    this.container = container;
    // this.dataManager =
    // session.getTransientNodesManager().getWorkspaceDataManager();
    // this.dataManager = (BatcheableWorkspaceDataManager) container
    // .getComponentInstanceOfType(BatcheableWorkspaceDataManager.class);
    this.name = name;
    this.observationManager = observationManager;
  }

  /**
   * @see javax.jcr.Workspace#getName
   */
  public String getName() {
    return name;
  }

  /**
   * @see javax.jcr.Workspace#importXML TODO (the same as Session.importXML - ?)
   */
  public ContentHandler getImportContentHandler(String parentAbsPath, int uuidBehavior)
      throws PathNotFoundException, ConstraintViolationException, VersionException,
      RepositoryException {
    NodeImporter contentHandler = (NodeImporter) session.getImportContentHandler(parentAbsPath,
        uuidBehavior);
    contentHandler.setSaveType(NodeImporter.SAVETYPE_SAVE);
    // Need to save all changes direct on SAX events
    //contentHandler.setSaveOnEnd(true);
    
    return contentHandler;
  }

  /**
   * @see javax.jcr.Workspace#importXML TODO (the same as Session.importXML - ?)
   */
  public void importXML(String parentAbsPath, InputStream in, int uuidBehavior) throws IOException,
      PathNotFoundException, ItemExistsException, ConstraintViolationException,
      InvalidSerializedDataException, RepositoryException {

    try {
      NodeImporter importer = (NodeImporter) getImportContentHandler(parentAbsPath, uuidBehavior);
      importer.parse(in);
    } catch (IOException e) {
      // e.printStackTrace();
      throw new InvalidSerializedDataException("importXML failed", e);
    } catch (SAXException e) {
      // e.printStackTrace();
      // throw new InvalidSerializedDataException("importXML failed", e);
      Throwable rootCause = e.getException();
      if (rootCause == null) {
        rootCause = session.getRootCauseException(e);
      }
      if (rootCause == null) {
        rootCause = e;
      }
      // rootCause.printStackTrace();
      if (rootCause instanceof ItemExistsException) {
        throw new ItemExistsException("importXML failed", rootCause);
      } else if (rootCause instanceof ConstraintViolationException) {
        throw new ConstraintViolationException("importXML failed", rootCause);
      } else {
        throw new InvalidSerializedDataException("importXML failed", e);
      }
    } catch (ParserConfigurationException e) {
      throw new InvalidSerializedDataException("importXML failed", e);
    }
  }

  /**
   * @see javax.jcr.Workspace#getSession
   */
  public Session getSession() {
    return session;
  }

  /**
   * @see javax.jcr.Workspace#getAccessibleWorkspaceNames
   */
  public String[] getAccessibleWorkspaceNames() throws RepositoryException {
    RepositoryImpl rep = (RepositoryImpl) session.getRepository();
    return rep.getWorkspaceNames();
  }

  /**
   * @see javax.jcr.Workspace#copy
   */
  public void copy(String srcWorkspace, String srcAbsPath, String destAbsPath)
      throws NoSuchWorkspaceException, ConstraintViolationException, VersionException,
      AccessDeniedException, PathNotFoundException, ItemExistsException, RepositoryException {

    // get source session
    SessionImpl srcSession = null;
    if (getName() != srcWorkspace) {
      srcSession = (SessionImpl) ((RepositoryImpl) session.getRepository()).login(session
          .getCredentials(), srcWorkspace);
    } else {
      srcSession = session;
    }

    // get destination node
    JCRPath destNodePath = session.getLocationFactory().parseAbsPath(destAbsPath);
    if (destNodePath.isIndexSetExplicitly())
      throw new RepositoryException(
          "The relPath provided must not have an index on its final element. "
              + destNodePath.getAsString(false));
    // get source node
    JCRPath srcNodePath = srcSession.getLocationFactory().parseAbsPath(srcAbsPath);
    NodeImpl srcNode = (NodeImpl) srcSession.getTransientNodesManager().getItem(
        srcNodePath.getInternalPath(), true);

    NodeImpl destParentNode = (NodeImpl) session.getTransientNodesManager().getItem(
        destNodePath.makeParentPath().getInternalPath(), true);
        

    if (srcNode == null || destParentNode == null) {
      throw new PathNotFoundException("No node exists at " + srcAbsPath
          + " or no node exists one level above " + destAbsPath);
    }
    try {
      destParentNode.checkPermission(PermissionType.ADD_NODE);
    } catch (AccessControlException e) {
      throw new AccessDeniedException(e.getMessage());
    }
    destParentNode.validateChildNode(destNodePath.getName().getInternalName(),
        ((ExtendedNodeType) srcNode.getPrimaryNodeType()).getQName());
    
    // Check for node with destAbsPath name in session
    NodeImpl destNode = (NodeImpl) session.getTransientNodesManager().getItem(
        destNodePath.getInternalPath(), true);

    if (destNode != null) {
      
      if (!destNode.getDefinition().allowsSameNameSiblings()) {
        // Throw exception
        String msg = "A node with name (" + destAbsPath + ") is already exists.";
        throw new ItemExistsException(msg);
      }
    }
    // Check if versionable Node is not checked-in
    if (!srcNode.isCheckedOut())
      throw new VersionException("Source parent node " + srcNode.getPath()
          + " or its nearest ancestor is checked-in");
    // Check locking
    if (!srcNode.checkLocking())
      throw new LockException("Source parent node " + srcNode.getPath() + " is locked ");
    
    ItemDataCopyVisitor initializer = new ItemDataCopyVisitor(
        (NodeData) destParentNode.getData(),
        destNodePath.getName().getInternalName(), 
        getNodeTypeManager(), 
        srcSession.getTransientNodesManager(), 
        false);
    srcNode.getData().accept(initializer);

    ItemDataChangesLog changesLog = new ItemDataChangesLog(initializer.getItemAddStates(), session.getId());
    
    session.getTransientNodesManager().getTransactManager().save(changesLog);
  }

  /**
   * @see javax.jcr.Workspace#copy
   */
  public void copy(String srcPath, String destPath) throws ItemExistsException, VersionException,
      PathNotFoundException, ItemExistsException, ConstraintViolationException, RepositoryException {

    copy(getName(), srcPath, destPath);
  }

  /**
   * @see javax.jcr.Workspace#move
   */
  public void move(String srcAbsPath, String destAbsPath) throws ConstraintViolationException,
      VersionException, AccessDeniedException, PathNotFoundException, ItemExistsException,
      RepositoryException {

    // get destination node
    JCRPath destNodePath = session.getLocationFactory().parseAbsPath(destAbsPath);
    if (destNodePath.isIndexSetExplicitly())
      throw new RepositoryException(
          "The relPath provided must not have an index on its final element. "
              + destNodePath.getAsString(false));
    // get source node
    JCRPath srcNodePath = session.getLocationFactory().parseAbsPath(srcAbsPath);
    NodeImpl srcNode = (NodeImpl) session.getTransientNodesManager().getItem(
        srcNodePath.getInternalPath(), true);

    NodeImpl destParentNode = (NodeImpl) session.getTransientNodesManager().getItem(
        destNodePath.makeParentPath().getInternalPath(), true);

    if (srcNode == null || destParentNode == null) {
      throw new PathNotFoundException("No node exists at " + srcAbsPath
          + " or no node exists one level above " + destAbsPath);
    }
    try {
      destParentNode.checkPermission(PermissionType.ADD_NODE);
      srcNode.checkPermission(PermissionType.REMOVE);
    } catch (AccessControlException e) {
      throw new AccessDeniedException(e.getMessage());
    }
    destParentNode.validateChildNode(destNodePath.getName().getInternalName(),
        ((ExtendedNodeType) srcNode.getPrimaryNodeType()).getQName());

    // Check for node with destAbsPath name in session
    NodeImpl destNode = (NodeImpl) session.getTransientNodesManager().getItem(
        destNodePath.getInternalPath(), true);

    if (destNode != null) {
      if (!destNode.getDefinition().allowsSameNameSiblings()) {
        // Throw exception
        String msg = "A node with name (" + destAbsPath + ") is already exists.";
        throw new ItemExistsException(msg);
      }
    }

    // Check if versionable ancestor is not checked-in
    if (!srcNode.isCheckedOut())
      throw new VersionException("Source parent node " + srcNode.getPath()
          + " or its nearest ancestor is checked-in");
    // Check locking
    if (!srcNode.checkLocking())
      throw new LockException("Source parent node " + srcNode.getPath() + " is     locked ");

    ItemDataMoveVisitor initializer = new ItemDataMoveVisitor((NodeData) destParentNode.getData(),
        destNodePath.getName().getInternalName(), getNodeTypeManager(), session
            .getTransientNodesManager(), srcNode.isNodeType(Constants.MIX_REFERENCEABLE));
    srcNode.getData().accept(initializer);

    // [PN] 06.01.07 Use one log - one transaction
    MutableItemDataChangesLog changes = new MutableItemDataChangesLog(session.getId());  
    // deleting nodes
//    ItemDataChangesLog changesDeletedLog = new ItemDataChangesLog(
//        initializer.getItemDeletedStates(true), session.getId());
    changes.addAll(initializer.getItemDeletedStates(true)); // changes.dump()
    
    // [PN] 06.01.07 Reindex same-name siblings on the parent after deletion
    changes.addAll(session.getTransientNodesManager().reindexSameNameSiblings(
        srcNode.nodeData(), session.getTransientNodesManager().getTransactManager()));
    
    // adding nodes
//    ItemDataChangesLog changesAddLog = new ItemDataChangesLog(
//        initializer.getItemAddStates(), session.getId());
    changes.addAll(initializer.getItemAddStates());
    
    session.getTransientNodesManager().getTransactManager().save(changes);
  }

  /**
   * @see javax.jcr.Workspace#clone
   */
  public void clone(String srcWorkspace, String srcAbsPath, String destAbsPath,
      boolean removeExisting) throws NoSuchWorkspaceException, ConstraintViolationException,
      VersionException, AccessDeniedException, PathNotFoundException, ItemExistsException,
      RepositoryException {

    if (srcWorkspace.equals(getName()))
      throw new RepositoryException("Source and destination workspace are equals " + name);

    // make dest node path
    JCRPath destNodePath = session.getLocationFactory().parseAbsPath(destAbsPath);

    if (destNodePath.isIndexSetExplicitly())
      throw new RepositoryException("DestPath should not contain an index " + destAbsPath);

    // find src node
    SessionImpl srcSession = (SessionImpl) ((RepositoryImpl) session.getRepository()).login(session
        .getCredentials(), srcWorkspace);

    // get source node
    JCRPath srcNodePath = srcSession.getLocationFactory().parseAbsPath(srcAbsPath);
    NodeImpl srcNode = (NodeImpl) srcSession.getTransientNodesManager().getItem(
        srcNodePath.getInternalPath(), true);

    NodeImpl destParentNode = (NodeImpl) session.getTransientNodesManager().getItem(
        destNodePath.makeParentPath().getInternalPath(), true);
    
    if (srcNode == null || destParentNode == null) {
      throw new PathNotFoundException("No node exists at " + srcAbsPath
          + " or no node exists one level above " + destAbsPath);
    }
    try {
      destParentNode.checkPermission(PermissionType.ADD_NODE);
    } catch (AccessControlException e) {
      throw new AccessDeniedException(e.getMessage());
    }
    
    destParentNode.validateChildNode(destNodePath.getName().getInternalName(),
        ((ExtendedNodeType) srcNode.getPrimaryNodeType()).getQName());
    
    // Check for node with destAbsPath name in session
    NodeImpl destNode = (NodeImpl) session.getTransientNodesManager().getItem(
        destNodePath.getInternalPath(), true);

    if (destNode != null) {
      if (!destNode.getDefinition().allowsSameNameSiblings()) {
        // Throw exception
        String msg = "A node with name (" + destAbsPath + ") is already exists.";
        throw new ItemExistsException(msg);
      }
    }

    // Check if versionable ancestor is not checked-in
    if (!srcNode.isCheckedOut())
      throw new VersionException("Source parent node " + srcNode.getPath()
          + " or its nearest ancestor is checked-in");
    // Check locking
    if (!srcNode.checkLocking())
      throw new LockException("Source parent node " + srcNode.getPath() + " is     locked ");

    ItemDataCloneVisitor initializer = new ItemDataCloneVisitor(
        (NodeData) destParentNode.getData(), destNodePath.getName().getInternalName(),
        getNodeTypeManager(), srcSession.getTransientNodesManager(), session
            .getTransientNodesManager(), removeExisting);
    srcNode.getData().accept(initializer);

    // [PN] to do save in one transaction
    MutableItemDataChangesLog changes = new MutableItemDataChangesLog(session.getId());
    
    // removeing existing nodes and properties
    if (removeExisting && initializer.getItemDeletedExistingStates(false).size()>0) {
//      ItemDataChangesLog changesDeletedExistingLog = new ItemDataChangesLog(initializer
//          .getItemDeletedExistingStates(true), session.getId());
//      session.getTransientNodesManager().getTransactManager().save(changesDeletedExistingLog);
      changes.addAll(initializer.getItemDeletedExistingStates(true));
    }

    // adding nodes
//    ItemDataChangesLog changesAddLog = new ItemDataChangesLog(initializer.getItemAddStates(),
//        session.getId());
    
    changes.addAll(initializer.getItemAddStates());
    
    session.getTransientNodesManager().getTransactManager().save(changes); // changes.dump()
  }

  /**
   * @see javax.jcr.Workspace#getQueryManager
   */
  public QueryManager getQueryManager() throws RepositoryException {
    QueryManagerFactory qf = (QueryManagerFactory) container
        .getComponentInstanceOfType(QueryManagerFactory.class);

    if (qf == null)
      throw new RepositoryException("Query Manager Factory not found. Check configuration.");
    else
      return qf.getQueryManager(session);
  }

  /**
   * @see javax.jcr.Workspace#getNodeTypeManager
   */
  public NodeTypeManagerImpl getNodeTypeManager() throws RepositoryException {
    return ((NodeTypeManagerImpl) container.getComponentInstanceOfType(NodeTypeManager.class))
        .createWorkspaceNTManager(session);
  }

  /**
   * @see javax.jcr.Workspace#getNamespaceRegistry
   */
  public NamespaceRegistry getNamespaceRegistry() throws RepositoryException {
    return (NamespaceRegistry) container.getComponentInstanceOfType(NamespaceRegistry.class);
  }

  /**
   * @see javax.jcr.Workspace#getObservationManager
   */
  public ObservationManager getObservationManager() throws UnsupportedRepositoryOperationException,
      RepositoryException {
    return observationManager;
  }

  private class MultiVersionRestoreComparator implements Comparator<Version> {
    public int compare(Version v1, Version v2) {
      return 0;
    }
  }

  /**
   * @see javax.jcr.Workspace#restore
   */
  public void restore(Version[] versions, boolean removeExisting)
      throws UnsupportedRepositoryOperationException, VersionException, RepositoryException,
      InvalidItemStateException {

    restoreVersions(versions, removeExisting);
    
//    if (session.hasPendingChanges())
//      throw new InvalidItemStateException("Session has pending changes ");
//
//    Exception versionException = null;
//    InternalQPath versionStorage = InternalQPath.parse(Constants.JCR_VERSION_STORAGE_URI);
//
//    // for restore operation
//    List<String> existedUuids = new ArrayList<String>(); // InWorkspace
//    List<Version> notExistedVersions = new ArrayList<Version>();
//    LinkedHashMap<Version, NodeImpl> existedVersions = new LinkedHashMap<Version, NodeImpl>();
//
//    for (Version v : versions) {
//      VersionHistory vh = v.getContainingHistory();
//      String versionableUuid = vh.getVersionableUUID();
//      NodeImpl node = null;
//      try {
//        node = (NodeImpl) session.getNodeByUUID(versionableUuid);
//        if (node.nodeData().getQPath().isDescendantOf(versionStorage, false)) {
//          continue; // skip version storage nodes (TODO:Remove in future)
//        }
//        existedVersions.put(v, node);
//        existedUuids.add(versionableUuid);
//      } catch (ItemNotFoundException e) {
//        // not found, looking for parent
//        // SPEC: For every version V in S that corresponds to a missing node in
//        // the workspace, there must
//        // also be a parent of V in S
//        // =========================================
//        // [PN], 14.02.06, So, we will play for TCK and will look at spec.
//        // Trying search for corresponding node, we hope that it's in
//        // VersionImpl object
//        // If we have a corr node and her parent in the existed list - all ok,
//        // otherwise exception will throws.
//        NodeImpl corrNode = null;
//        String versionableParentUuid = null;
//        if (!v.getSession().getWorkspace().getName().equals(session.getWorkspace().getName())) {
//          try {
//            corrNode = (NodeImpl) v.getSession().getNodeByUUID(versionableUuid);
//            if (corrNode.nodeData().getQPath().isDescendantOf(versionStorage, false)) {
//              continue; // skip version storage nodes (TODO:Remove in future)
//            }
//            versionableParentUuid = corrNode.getParent().getProperty("jcr:uuid").getString();
//          } catch (ItemNotFoundException e1) {
//            // not found in version session
//            log.warn("Workspace.restore(): corr node not found: " + versionableUuid, e);
//            versionException = e;
//          } catch (RepositoryException e1) {
//            log.warn("Workspace.restore(): Repository error, for corr node: " + versionableUuid, e);
//            versionException = e;
//          }
//        }
//        if (versionableParentUuid != null && existedUuids.contains(versionableParentUuid)) {
//          notExistedVersions.add(v);
//          continue;
//        } else {
//          throw new VersionException("No such node (for version, from the array of versions, "
//              + "that corresponds to a missing node in the workspace, "
//              + "there must also be a parent in the array). UUID: " + versionableUuid,
//              versionException);
//        }
//      }
//    }
//
//    for (Version v : existedVersions.keySet()) {
//      String versionableUuid = v.getContainingHistory().getVersionableUUID();
//      // [PN], 14.02.06, 20.02.06
//      try {
//        NodeImpl node = existedVersions.get(v);
//        ((VersionImpl) v).restore(node, removeExisting);
//      } catch (ItemNotFoundException e) {
//        throw new VersionException("No such node. UUID: " + versionableUuid, e);
//      } catch (ItemExistsException e) {
//        log.warn("Workspace.restore(): " + e.getMessage());
//        throw e;
//      } catch (RepositoryException e) {
//        log.warn("Workspace.restore(): Repository error: " + versionableUuid, e);
//        throw e;
//      }
//    }
//    for (Version v : notExistedVersions) {
//      String versionableUuid = v.getContainingHistory().getVersionableUUID();
//      NodeImpl node = null;
//      try {
//        node = (NodeImpl) session.getNodeByUUID(versionableUuid);
//        ((VersionImpl) v).restore(node, removeExisting);
//      } catch (ItemNotFoundException e) {
//        throw new VersionException(
//            "No such node restored before (for version, from the array of versions, "
//                + "that corresponds to a missing node in the workspace, "
//                + "there must also be a parent in the array). UUID: " + versionableUuid);
//      } catch (ItemExistsException e) {
//        log.warn("Workspace.restore() for item not existed before: " + e.getMessage());
//        throw e;
//      } catch (RepositoryException e) {
//        log.warn("Workspace.restore() for item not existed before: Repository error: "
//            + versionableUuid, e);
//        throw e;
//      }
//    }
  }
  
  protected void restoreVersions(Version[] versions, boolean removeExisting)
      throws UnsupportedRepositoryOperationException, VersionException, RepositoryException,
      InvalidItemStateException {

    if (session.hasPendingChanges())
      throw new InvalidItemStateException("Session has pending changes ");

    // for restore operation
    List<String> existedUuids = new ArrayList<String>(); // InWorkspace
    List<VersionImpl> notExistedVersions = new ArrayList<VersionImpl>();
    LinkedHashMap<VersionImpl, NodeData> existedVersions = new LinkedHashMap<VersionImpl, NodeData>();

    DataManager dataManager = session.getTransientNodesManager().getTransactManager();

    for (Version v : versions) {
      String versionableUuid = v.getContainingHistory().getVersionableUUID();
      NodeData node = (NodeData) dataManager.getItemData(versionableUuid);
      if (node != null) {
        existedVersions.put((VersionImpl) v, node);
        existedUuids.add(versionableUuid);
      } else {
        // not found, looking for parent
        // SPEC: For every version V in S that corresponds to a missing node in
        // the workspace, there must also be a parent of V in S
        // =========================================
        // Trying search for corresponding node,
        // If we have a corr node and her parent in the existed list - all ok,
        // otherwise exception will be thrown.
        NodeData corrNode = null;
        String versionableParentUuid = null;
        if (!v.getSession().getWorkspace().getName().equals(session.getWorkspace().getName())) {
          DataManager vDataManager = ((SessionImpl) v.getSession()).getTransientNodesManager().getTransactManager();
          corrNode = (NodeData) vDataManager.getItemData(versionableUuid);
          if (corrNode != null)
            versionableParentUuid = corrNode.getParentUUID();
          else
            log.warn("Workspace.restore(). Correspondent node is not found " + versionableUuid);
        }
        if (versionableParentUuid != null && existedUuids.contains(versionableParentUuid)) {
          notExistedVersions.add((VersionImpl) v);
          continue;
        }
        throw new VersionException("No such node (for version, from the array of versions, "
            + "that corresponds to a missing node in the workspace, "
            + "there must also be a parent in the array). UUID: " + versionableUuid);

      }
    }

    SessionChangesLog changesLog = new SessionChangesLog(session.getId());

    for (VersionImpl v : existedVersions.keySet()) {
      try {
        NodeData node = existedVersions.get(v);
        VersionHistoryDataHelper historyData = v.getContainingHistory().getData();

        changesLog.addAll(v.restoreLog(node, historyData, session, removeExisting, changesLog).getAllStates());
      } catch (ItemExistsException e) {
        throw new ItemExistsException("Workspace restore. Can't restore a node. "
            + v.getContainingHistory().getVersionableUUID() + ". " + e.getMessage(), e);
      } catch (RepositoryException e) {
        throw new RepositoryException("Workspace restore. Can't restore a node. "
            + v.getContainingHistory().getVersionableUUID() + ". Repository error: "
            + e.getMessage(), e);
      }
    }

    // MultiVersionRestoreComparator<Version> orderer = new
    // MultiVersionRestoreComparator<Version>();
    // Collections.sort(notExistedVersions, orderer);

    for (VersionImpl v : notExistedVersions) {
      String versionableUuid = v.getContainingHistory().getVersionableUUID();
      try {
        NodeData node = null;
        for (ItemState change : changesLog.getAllStates()) {
          if (change.isNode() && change.isAdded()
              && ((NodeData) change.getData()).getUUID().equals(versionableUuid)) {
            node = (NodeData) change.getData();
            break;
          }
        }
        if (node != null) {
          VersionHistoryDataHelper historyData = v.getContainingHistory().getData();
          changesLog.addAll(v.restoreLog(node, historyData, session, removeExisting, changesLog).getAllStates());
        } else {
          throw new VersionException(
              "No such node restored before (for version, from the array of versions, "
                  + "that corresponds to a missing node in the workspace, "
                  + "there must also be a parent in the array). UUID: " + versionableUuid);
        }
      } catch (ItemExistsException e) {
        throw new ItemExistsException(
            "Workspace restore. Can't restore a node not existed before. " + versionableUuid + ". "
                + e.getMessage(), e);
      } catch (RepositoryException e) {
        throw new RepositoryException(
            "Workspace restore. Can't restore a node not existed before. " + versionableUuid
                + ". Repository error: " + e.getMessage(), e);
      }
    }

    dataManager.save(changesLog);
  }
  
}