/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.xml;

import java.util.Collections;
import java.util.List;

import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.exoplatform.services.jcr.dataflow.ItemDataConsumer;
import org.exoplatform.services.jcr.dataflow.ItemDataTraversingVisitor;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.dataflow.NodeDataOrderComparator;
import org.exoplatform.services.jcr.impl.dataflow.PropertyDataOrderComparator;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

/**
 * @author <a href="mailto:Sergey.Kabashnyuk@gmail.com">Sergey Kabashnyuk</a>
 * @version
 */
public abstract class ExportXmlVisitor extends ItemDataTraversingVisitor {

  /**
   * Actual value of each <code>PropertyType.BINARY</code> property is
   * recorded using Base64 encoding
   */
  public static final int      BINARY_PROCESS = 0;

  /**
   * Any properties of <code>PropertyType.BINARY</code> will be ignored and
   * will not appear in the serialized output
   */
  public static final int      BINARY_SKIP    = 1;

  /**
   * Serialized only definition of of each <code>PropertyType.BINARY</code>
   * property without values
   */
  public static final int      BINARY_EMPTY   = 2;

  private int                  binaryConduct  = BINARY_PROCESS;

  protected boolean            noRecurse;

  private final ContentHandler contentHandler;

  protected final Session session;

  protected LocationFactory locationFactory;

  public ExportXmlVisitor(ContentHandler handler, Session session, ItemDataConsumer dataManager,
      boolean skipBinary, boolean noRecurse) {

    super(dataManager, noRecurse ? 1 : -1);
    this.session = session;
    if (session instanceof SessionImpl) {
      this.locationFactory = ((SessionImpl) session).getLocationFactory();
    } else {
      throw new java.lang.IllegalArgumentException("session not an instanceof SessionImpl");
    }

    this.contentHandler = handler;
    setBinaryConduct(skipBinary ? BINARY_SKIP : BINARY_PROCESS);
    setNoRecurse(noRecurse);
  }

  /**
   * Specify how properties of <code>PropertyType.BINARY</code> serialized
   * 
   * @see <code>BINARY_PROCESS</code>,<code>BINARY_SKIP</code>,
   *      <code>BINARY_EMPTY</code>
   */
  public int getBinaryConduct() {
    return binaryConduct;
  }

  public void setBinaryConduct(int binaryConduct) {
    if (binaryConduct != BINARY_PROCESS && binaryConduct != BINARY_SKIP
        && binaryConduct != BINARY_EMPTY)
      throw new java.lang.IllegalArgumentException("binaryConduct must be one of "
          + "BINARY_PROCESS,BINARY_SKIP, BINARY_EMPTY");
    this.binaryConduct = binaryConduct;
  }

  public boolean isNoRecurse() {
    return noRecurse;
  }

  public void setNoRecurse(boolean noRecurse) {
    this.noRecurse = noRecurse;
  }

  public void export(NodeData node) throws RepositoryException, SAXException {
    if (contentHandler != null) {
      contentHandler.startDocument();
      startPrefixMapping();
      node.accept(this);
      endPrefixMapping();
      contentHandler.endDocument();
    }
  }

 protected void startPrefixMapping() throws RepositoryException, SAXException{
   String[] prefixes = session.getNamespacePrefixes();
   for (String prefix : prefixes) {
     contentHandler.startPrefixMapping(prefix, session.getNamespaceURI(prefix));
   }

 };

  protected void endPrefixMapping() throws RepositoryException, SAXException{
    String[] prefixes = session.getNamespacePrefixes();
    for (String prefix : prefixes) {
      contentHandler.endPrefixMapping(prefix);
    }
  }
    ;

  @Override
  public void visit(NodeData node) throws RepositoryException {
    try {
      entering(node, currentLevel);
      if (maxLevel == -1 || currentLevel < maxLevel) {
        currentLevel++;

        List<PropertyData> properies = dataManager.getChildPropertiesData(node);
        // Sorting properties
        Collections.sort(properies, new PropertyDataOrderComparator());
        for (PropertyData data : properies)
          data.accept(this);
        
        if (!isNoRecurse() && currentLevel > 0) {
          List<NodeData> nodes = dataManager.getChildNodesData(node);
          // Sorting nodes
          Collections.sort(nodes, new NodeDataOrderComparator());
          for (NodeData data : nodes)
            data.accept(this);
        }
        currentLevel--;
      }
      leaving(node, currentLevel);
    } catch (RepositoryException re) {
      currentLevel = 0;
      throw re;
    }

  }

}
