package org.esupportail.cas.server;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Date;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.URLDecoder;

import javax.servlet.ServletRequest;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.dom4j.XPath;
import org.esupportail.cas.server.util.BasicHandler;
import org.esupportail.cas.server.util.MisconfiguredHandlerException;
import org.esupportail.cas.server.util.RedundantHandler;
import org.esupportail.cas.server.util.Server;
import org.esupportail.cas.server.util.log.Debug;
import org.esupportail.cas.server.util.log.Log;

import edu.yale.its.tp.cas.auth.provider.WatchfulPasswordHandler;


//---------------------------------------------------------

/**
 * This class permits via a Xml configuration file to call different specific handler
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 * @author Jean-Baptiste Daniel <danielj at sourceforge.net>
 */

public final class GenericHandler extends WatchfulPasswordHandler {    
	
	/**
	 * the package release number.
	 */
	private static final String CASGENERICHANDLER_RELEASE = "2.0.5-1";
	
	/**
	 * Debugging mode.
	 */
	private static boolean debug = false;
	
	/**
	 * the list of all the specific handlers to try for authentication.
	 */
	private static List handlers = null;
	
	/**
	 * the name of the configuration file.
	 */
	private static String configFilename = null;
	
	/**
	 * a boolean to store wheter empty passwords can be accepted or not.
	 */
	private static boolean acceptEmptyPasswords;
	
	/**
	 * Retrieve the name of the file storing the configuration.
	 *
	 * @return a string.
	 */
	private static String getConfigFileName() {
		java.net.URL resourceURL;

		if (configFilename == null) {
			resourceURL = GenericHandler.class.getResource("/../genericHandler.xml");
			if (resourceURL == null) {
				Log.warn("Configuration file genericHandler.xml is missing!");
			}
			// if configuration file is missing an exception is thrown and the cas server crashes
			configFilename = resourceURL.getFile();

			// decode string to allow paths with spaces (since version 2.0.4)
			// J-Louis.RENAUD at univ-bpclermont.fr
            try {
                configFilename = URLDecoder.decode(configFilename, "UTF-8");
            } catch (UnsupportedEncodingException e) {
				Log.warn("Configuration filename could not be decoded!");
            }
		}
		return configFilename;
	}
	
	/**
	 * the last time the configuration file was modified
	 */
	private static long configDate = 0;
	
	/**
	 * tell if the configuration file was modified.
	 *
	 * @return true if the file was modified since last configuration reading.
	 */
	private static boolean configFileModified() {
		return new File(getConfigFileName()).lastModified() > configDate;
	}
	
	/**
	 * Scans an XML configuration file and builds the structure needed 
	 * for authentication (a list of specific handlers).
	 *
	 * @throws MisconfiguredHandlerException MisconfiguredHandlerException
	 */
	private void readConfigFile() throws MisconfiguredHandlerException {
		
		// create an empty list of handlers
		List newHandlers = new LinkedList();
		
		Document config;
		XPath xPathSelector;
		Element authenticationElement;
		
		// create a SAXReader component
		SAXReader saxReader = new SAXReader(false);
		
		// parse the configuration file
		try {
			config = saxReader.read(getConfigFileName());
		} catch (DocumentException e) {
			throw new MisconfiguredHandlerException("Configuration file is ill-formed: " + e.getMessage());
		}
		
		// check the root node
		xPathSelector = DocumentHelper.createXPath("/authentication");
		try {
			authenticationElement = (Element) xPathSelector.selectNodes(config).get(0);
		} catch (IndexOutOfBoundsException e) {
			throw new MisconfiguredHandlerException("Root <authentication> tag not found.");
		}
		boolean configDebug = Debug.elementDebugValue(authenticationElement, false);
		if (configDebug) {
			Log.debug("Debugging mode set to \"on\".");
		}
		boolean newAcceptEmptyPasswords = authenticationElement.attributeValue(
				"empty_password_accepted", 
				"off").equals("on");
		if (configDebug) {
			if (newAcceptEmptyPasswords) {
				Log.debug("Empty passwords will be accepted.");
			} else {
				Log.debug("Empty passwords will not be accepted.");
			}
		}
		// get the list of all the handlers		
		xPathSelector = DocumentHelper.createXPath("/authentication/*");
		List handlerElements = xPathSelector.selectNodes(config);
		
		// check that at least one handler is declared 
		if (handlerElements.isEmpty()) {
			throw new MisconfiguredHandlerException("No \"handler\" element found.");
		}

		for (Iterator i = handlerElements.iterator(); i.hasNext();) {
			Element handlerElement = (Element) i.next();
			
			// check that all the sub-elements are handlers
			if (!handlerElement.getName().equals("handler")) {
				throw new MisconfiguredHandlerException(
						"Unknown \"" 
						+ handlerElement.getName() 
						+ "\" XML element found.");
			}
			if (configDebug) { Log.debug("Found a handler."); }
			
			// extract the classname
			Element handlerClassnameElement = handlerElement.element("classname");
			if (handlerClassnameElement == null) {
				throw new MisconfiguredHandlerException(
						"No \"classname\" element found for a handler.");
			}
			String handlerClassname = handlerClassnameElement.getTextTrim();
			if (handlerClassname.equals("")) {
				throw new MisconfiguredHandlerException(
						"Empty \"classname\" element found for a handler.");
			}
			if (configDebug) { Log.debug("This is a \"" + handlerClassname + "\" handler."); }

			// try to load the class
			Class handlerClass;
			try {
				handlerClass = Class.forName(handlerClassname);
			} catch (ClassNotFoundException e) {
				throw new MisconfiguredHandlerException(
						"Class \"" + handlerClassname + "\" could not be loaded.");
			}

			// get the constructor
			Class[] handlerConstructorArgumentTypes = {Element.class, Boolean.class};
			Constructor handlerConstructor;
			try {
				handlerConstructor = handlerClass.getConstructor(handlerConstructorArgumentTypes);
			} catch (Exception e) {
				throw new MisconfiguredHandlerException(
						"The constructor of Class \"" 
						+ handlerClassname 
						+ "\" could not be loaded because a \"" 
						+ e.getClass().getName() 
						+ "\" exception was raised: " 
						+ e.getMessage());
			}

			// create an instance
			Object[] handlerConstructorArguments = { 
					handlerElement, 
					new Boolean(configDebug) };
			BasicHandler handler;
			try {
				handler = (BasicHandler) handlerConstructor.newInstance(handlerConstructorArguments);
			} catch (InvocationTargetException e) {
				throw new MisconfiguredHandlerException(e.getCause().getMessage());
			} catch (Exception e) {
				throw new MisconfiguredHandlerException(
						"Class \"" 
						+ handlerClassname 
						+ "\" could not be instanciated because a \"" 
						+ e.getClass().getName() 
						+ "\" exception was raised: " 
						+ e.getMessage());
			}

			newHandlers.add(handler);
		}
		
		handlers = newHandlers;
		debug = configDebug;
		acceptEmptyPasswords = newAcceptEmptyPasswords;
		
		// while we created handlers and servers, we set the debugging mode
		// on each object depending on the confifuration. Now we set debugging mode
		// to on for redundant handlers having servers with debugging mode set
		// to on, and then the main handler if one specific handler (at least) has
		// debugging mode set to on.
		for (Iterator i = handlers.iterator(); i.hasNext();) {
			BasicHandler handler = (BasicHandler) i.next();
			if (handler instanceof RedundantHandler) {
				for (Iterator j = ((RedundantHandler) handler).getServers().iterator(); j.hasNext();) {
					Server server = (Server) j.next();
					if (server.isDebug()) {
						handler.setDebug(true);
					}
				}
			}
			if (handler.isDebug()) {
				debug = true;
			}
		}
		
		// configuration read without any error
		configDate = new Date().getTime();
	}    

	/**
	 * Try to authenticate a user by calling all the handlers found in
	 * the configuration file.
	 *
	 * @param request the current request
	 * @param username the username provided by the user
	 * @param password the password provided by the user
	 *
	 * @return true on success, false otherwise.
	 */
	public synchronized boolean authenticate(
			final ServletRequest request, 
			final String username, 
			final String password) {
		if (!super.authenticate(request, username, password)) {
			Log.warn("Authentication process was blocked by WatchfullPasswordHandler.");
			return false;
		}
		
		if (configFileModified()) {
			Log.info(new String(
					"ESUP-Portail Generic Handler " 
					+ CASGENERICHANDLER_RELEASE 
					+ ", reading configuration file..."));			
			try {
				readConfigFile();
				Log.info("Configuration file read without any error.");
			} catch (MisconfiguredHandlerException e) {
				Log.error("An error occured while reading the configuration file:");
				Log.error(e.getMessage());
				if (handlers == null) {
					Log.warn("No previous configuration, authentication will always fail.");
				} else {
					Log.warn("Keeping the previous configuration.");
				}
			}
		}	

		// as the debug flag is set by readConfigFile(), a trace can be written only at this point
		if (debug) { Log.traceBegin(); }
		
		if (handlers == null) {
			Log.warn("No authentication handler found.");
		}
		
		if (username.equals("")) {
			if (debug) { Log.debug("Empty username, handlers will not be called."); }
		} else if (password.equals("") && !acceptEmptyPasswords) {
			if (debug) { Log.debug("Empty password, handlers will not be called."); }
		} else if (handlers != null) {
			for (Iterator i = handlers.iterator(); i.hasNext();) {
				BasicHandler handler = (BasicHandler) i.next();
				switch (handler.authenticate(username, password)) {
					case BasicHandler.SUCCEEDED:
						Log.info("Authentication succeeded for user `" + username + "'.");
						if (debug) { Log.traceEnd(String.valueOf(true)); }
						return true;
					case BasicHandler.FAILED_STOP:
						Log.info("Authentication failed for user `" + username + "'.");
						if (debug) { Log.traceEnd(String.valueOf(false)); }
						return false;
					default: // BasicHandler.FAILED_CONTINUE
				}
			}
		}
		Log.info("Authentication failed for user `" + username + "'.");
		if (debug) { Log.traceEnd(String.valueOf(false)); }
		return false;
	}
	
	
}
