package org.esupportail.cas.server.handlers.file;

import java.io.FileNotFoundException;

import java.io.BufferedReader;
import java.io.FileReader;

import org.dom4j.Element;
import org.esupportail.cas.server.util.BasicHandler;
import org.esupportail.cas.server.util.MisconfiguredHandlerException;
import org.esupportail.cas.server.util.crypt.Crypt;
import org.esupportail.cas.server.util.log.Log;

/**
 * The specific Handler for file.
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 * @author Jean-Baptiste Daniel <danielj at sourceforge.net>
 */
public final class FileHandler extends BasicHandler {
	
	/**
	 * The filename to read.
	 */
	private String filename;
	/**
	 * The character used to sperate fields in the file.
	 */
	private String separator;
	/**
	 * the encryption used to store passwords.
	 */
	private String encryption;

	/**
	 * Constructor.
	 *
	 * @param handlerElement the XML element that declares the handler 
	 * in the configuration file
	 * @param configDebug debugging mode of the global configuration
	 * @throws Exception Exception
	 */
	public FileHandler(
			final Element handlerElement, 
			final Boolean configDebug) throws Exception {
		super(handlerElement, configDebug);
		traceBegin();

		checkConfigElement(true);

		filename = getConfigSubElementContent("filename", true/*needed*/);
		trace("filename = " + filename);
		
		encryption = getConfigSubElementContent("encryption", false/*not needed*/);
		if (encryption.equals("")) {
			encryption = "md5";
		}
		if (!Crypt.isEncryptionSupported(encryption)) {
			traceThrow(new MisconfiguredHandlerException(
					"Encryption \"" 
					+ encryption 
					+ "\" is not supported."));
		}
		trace("encryption = " + encryption);
		if (encryption.equals("plain")) {
			Log.warn("Passwords should be encrypted. Be sure to keep " + filename + " out of danger!");
		}

		separator = getConfigSubElementContent("separator", false/*not needed*/);
		trace("separator = " + separator);

		traceEnd();
	}
	
	/**
	 * Try to authenticate a user (by searching into a file).
	 *
	 * @param username the user's name
	 * @param password the user's password
	 *
	 * @return BasicHandler.SUCCEDED on success, or 
	 * BasicHandler.FAILED_CONTINUE otherwise.
	 */	
	public int authenticate(final String username,
			final String password) {
		traceBegin();
		
		try {
			trace("Opening file...");
			FileReader fileReader = new FileReader(filename);
			trace("Creating a buffer...");
			BufferedReader buf = new BufferedReader(fileReader);
			
			String line;
			
			trace("Reading file...");
			while ((line = buf.readLine()) != null) {
				String[] lineFields = line.split(separator);
				
				// compare the username
				if (lineFields[0].equals(username)) {
					
					trace("Username found, checking password (" + encryption + ")...");
					String encryptedPassword = lineFields[1];
					boolean match = Crypt.match(encryption, password, encryptedPassword);
					
					// function will return now (even if passwords do not match
					buf.close();
					
					// compare the password
					if (match) {
						trace("Password matches.");
						traceEnd("SUCCEEDED");
						return SUCCEEDED;
					} else {
						trace("Password does not match.");
						traceEnd("FAILED_CONTINUE");
						return FAILED_CONTINUE;
					}
				}
			}
			buf.close();
			trace("Username not found.");
		} catch (FileNotFoundException e) {
			Log.warn("File \"" + filename + "\" could not be loaded.");
			trace("Failure: " + e.toString());
		} catch (Exception e) {
			trace("Failure: " + e.toString());
		}
		traceEnd("FAILED_CONTINUE");
		return FAILED_CONTINUE;
	}
}
