package org.esupportail.cas.server.handlers.ldap;

import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.AuthenticationException;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;

import org.dom4j.Element;
import org.esupportail.cas.server.util.log.Log;
import org.esupportail.cas.server.util.RedundantHandler;
import org.esupportail.cas.server.util.Server;

/**
 * This abstract class implements an LDAP server class, inherited by
 * BindLdapServer and FastBindLdapServer. 
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 * @author Jean-Baptiste Daniel <danielj at sourceforge.net>
 */
public abstract class LdapServer extends Server {

	/**
	 * true for a secure access to the LDAP directory, false otherwise.
	 */
	private boolean secured;
	
	/**
	 * the URL of the LDAP directory.
	 */
	private String url;
	
	/**
	 * Constructor.
	 *
	 * @param handlerDebug debugging mode of the handler
	 * @param handler      the handler the server will be used by
	 * @param serverElement the XML element that declares the server 
	 * @throws Exception Exception
	 */
	public LdapServer(
			final Boolean handlerDebug,
			final RedundantHandler handler,
			final Element serverElement) throws Exception {
		super(handlerDebug, handler, serverElement);
		traceBegin();

		url = getServerSubElementContent(serverElement, "url", true/*needed*/);
		trace("url = " + url);

		secured = url.substring(0, 5).equals("ldaps");
		trace("secured = " + secured);

		traceEnd();
	}
	
	/**
	 * A String array used to store input tokens (to be replaced). 
	 */
	private static final String[] INPUT_TOKENS = { 
			"%u", "%U", "%d", "%1", "%2", "%3", "%4", "%5", "%6", "%7", "%8", "%9" 
	};
	
	/**
	 * Take a filter (from CASGenericHandler configuration) and a
	 * username, and replace tokens in the filter by their equivalents 
	 * in the username. The rules to replace tokens are the following ones:
	 * - %%: %
	 * - %u: user
	 * - %U: user portion of %u (%U = test when %u = test@domain.tld)
	 * - %d: domain portion of %u (%d = domain.tld when %u = test@domain.tld)
	 * - %1-9: domain tokens (%1 = tld, %2 = domain when %d = domain.tld)
	 *
	 * @param username a username (such as test@domain.com)
	 * @param filter a filter string containing tokens to be replaced
	 *
	 * @return a string corresponding to the input filter, where the
	 * predefined tokens have been replaced by their equivalents.
	 */
	protected final String replaceTokens(final String filter,
			final String username) {
		traceBegin();
		
		/* A String array used to store output tokens (to replace
		 * input tokens in the replaceTokens() method). */
		String[] outputTokens = { 
				"", "", "", "", "", "", "", "", "", "", "", "", "" 
		};
		
		/* -------------------------------------------------------------
		 * Analyze the username given by the user and deduce the output
		 * tokens (which will replace the input tokens when calling 
		 * replaceTokens()).
		 * ------------------------------------------------------------- */
		
		// the complete username corresponds to '%u', store it
		outputTokens[0] = username;
		
		// at first cut the username into two parts, seperated by a '@'
		String[] userDomain = username.split("@");
		// the first cell of userDomain corresponds to '%U', store it
		outputTokens[1] = userDomain[0];
		
		// if a '@' it present,
		if (userDomain.length > 1) {
			
			//the second cell corresponds to '%d', store it
			outputTokens[2] = userDomain[1];
			
			// now look at the '%n' tokens. Split the domain name into pieces
			String[] dcArray = userDomain[1].split("\\.");
			
			// iterate on the array to fill the end of _outputTokens
			for (int i = 0; i < dcArray.length; i++) {
				outputTokens[i + 3] = dcArray[dcArray.length - 1 - i];
			}
		}
		
		/* -------------------------------------------------------------
		 * Replace the input tokens by their corresponding parts.
		 * ------------------------------------------------------------- */
		StringBuffer result = new StringBuffer("");
		
		// at first split _username into parts separated by "%%"
		String[] parts = filter.split("%%");
		
		// for each part, replace the tokens and concatenate to result
		for (int i = 0; i < parts.length; i++) {
			if (i != 0) {
				result.append("%");
			}
			for (int j = 0; j < INPUT_TOKENS.length; j++) {
				parts[i] = parts[i].replaceAll(INPUT_TOKENS[j], outputTokens[j]);
			}
			result.append(parts[i]);
		}
		traceEnd(result.toString());
		return result.toString();
	}
	
	/**
	 * Connect to the LDAP server using specified username and password.
	 *
	 * @param bindDn       the DN to use for the connection
	 * @param bindPassword the associated password
	 *
	 * @return a Connection object on success, null on error. When the function 
	 * returns false, the error code can be retrieved with the connectError() 
	 * method.
	 */	
	protected final DirContext connect(final String bindDn,
									   final String bindPassword) {
		DirContext connection = null;
		traceBegin();
		
		try {
			Hashtable hashtable = new Hashtable(5, 0.75f);
			hashtable.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
			hashtable.put(Context.PROVIDER_URL, url);
			hashtable.put(Context.SECURITY_AUTHENTICATION, "simple");
			hashtable.put(Context.SECURITY_PRINCIPAL, bindDn);
			hashtable.put(Context.SECURITY_CREDENTIALS, bindPassword.getBytes());
			if (secured) {
				hashtable.put(Context.SECURITY_PROTOCOL, "ssl");
			}
			trace("Connecting to the LDAP directory (url=`" + url + "', username=`" + bindDn + "')...");
			connection = new InitialDirContext(hashtable);
			setConnectError(CONNECT_SUCCESS);
			trace("Connection succeeded.");
			traceEnd("ok");
			return connection;
		} catch (AuthenticationException e) {
			trace("Connection failed: " + e.getMessage());
			setConnectError(CONNECT_NOAUTH);
		} catch (NamingException e) {
			Log.warn("Could not connect to \"" + url + "\"");
			trace("Connection failed: '" + e.getMessage() + "')");
			setConnectError(CONNECT_FAILURE);
		}

		// connection failed, but try to close the connection however
		if (connection != null) {
			try {
				trace("Closing LDAP connection...");
				connection.close();
			} catch (NamingException e2) {
				Log.warn("Could not close LDAP connection.");
			}
		}

		traceEnd("null");
		return null;
	}
	
	/**
	 * Connect to the LDAP server using specified username and password
	 * and immediatly close the connection. The error code can be retrieved 
	 * with the connectError() method.
	 *
	 * @param bindDn       the DN to use for the connection
	 * @param bindPassword the associated password
	 * 
	 * @return true if the connection succeeded, false otherwise.
	 */	
	protected final boolean connectAndClose(
			final String bindDn,
			final String bindPassword) {
		traceBegin();
		
		DirContext connection = connect(bindDn, bindPassword);
		if (connection != null) {
			try {
				trace("Closing LDAP connection...");
				connection.close();
				return true;
			} catch (NamingException e) {
				Log.warn("Could not close LDAP connection.");
				return false;
			}
		} else {
			return false;
		}
	}
}

