package org.esupportail.cas.server.util.log;

import org.dom4j.Element;
import org.esupportail.cas.server.util.MisconfiguredHandlerException;

/**
 * This class implements a basic object owning debugging features.
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 */
public abstract class Debug {
	/**
	 * A debugging flag.
	 */
	private boolean debug;
	
	/**
	 * Tell if the handler is in debugging mode or not.
	 *
	 * @return true when the handler is in debugging mode, false otherwise.
	 */
	public final boolean isDebug() {
		return debug;
	}
	
	/**
	 * Set deubgging mode.
	 *
	 * @param d a boolean
	 */
	public final void setDebug(final boolean d) {
		this.debug = d;
	}
	
	/** 
	 * Debugging. 
	 * @param str a string 
	 */
	protected final void trace(final String str) { 
		if (debug) {
			Log.debug(str); 
		}
	}
	/** 
	 * Debugging. 
	 * @param obj an object 
	 */
	protected final void trace(final Object obj) { 
		trace(obj.toString());
	}

	/** 
	 * Debugging (at the beginning of the execution of a method). 
	 */
	protected final void traceBegin() { 
		if (debug) {
			Log.traceBegin(); 
		}
	}
	/** 
	 * Debugging (at the end of the execution of a method). 
	 * @param str a string 
	 */
	protected final void traceEnd(final String str) { 
		if (debug) {
			Log.traceEnd(str); 
		}
	}
	/** 
	 * Debugging (at the end of the execution of a method). 
	 * @param obj an object 
	 */
	protected final void traceEnd(final Object obj) { 
		traceEnd(obj.toString()); 
	}
	/** Debugging (at the end of the execution of a method). */
	protected final void traceEnd() { 
		traceEnd(""); 
	}
	
	/** 
	 * Debugging (throwing an exception).
	 * 
	 * @param e the exception to throw
	 * @throws Exception Exception
	 */
	protected final void traceThrow(final Exception e) throws Exception { 
		traceEnd("THROWS " + e.getClass().getName() + ": " + e.getMessage());
		throw e;
	}
	
	/**
	 * Constructor.
	 *
	 * @param d true to set debugging mode, false otherwise.
	 */
	protected Debug(final boolean d) {
		super();
		debug = d;
		traceBegin();
		trace("debug = " + debug);
		traceEnd();
	}

	/**
	 * A convenient method to get debugging mode with a default value.
	 *
	 * @param element an XML element having a debug attribute ("on", "off" or "default")
	 * @param defaultValue the value to return when str equals to "default"
	 *
	 * @return a boolean.
	 */
	public static boolean elementDebugValue(
			final Element element,
			final boolean defaultValue) {
		String str = element.attributeValue("debug", "default");
		if (str.equals("on")) {
			return true;
		} else if (str.equals("off")) {
			return false;
		} else {
			return defaultValue;
		}
	}
	
	/**
	 * Check that a class is present
	 * 
	 * @param classname the name of the class
	 * @return the corresponding class
	 * @throws Exception Exception
	 */
	protected final Class checkClass(final String classname) throws Exception {
		traceBegin();
		trace("Checking if class \"" + classname + "\" is available...");
		try {
			traceEnd("ok");
			return Class.forName(classname);
		} catch (ClassNotFoundException e) {
			trace("Class not found.");
			traceThrow(new MisconfiguredHandlerException(
					"Class \"" + classname + "\" is needed by \"" 
					+ getClass().getName() 
					+ "\" handlers."
					));
			// never reached
			return null;
		}
	}
	
}

